/* Mex interface to David MacKay's macopt gradient-only optimizer.
 * Calling documentation in macopt.m
 * 
 * More information about the macopt optimiser itself is here:
 * 	http://www.inference.phy.cam.ac.uk/mackay/c/macopt.html
 * 	mackay@mrao.cam.ac.uk
 * Note that the optimizer is GPL software, please look at that code for
 * licenses and copyright notices.
 * 
 * This mex interface by Iain Murray, http://www.gatsby.ucl.ac.uk/~iam23/
 * June 2004.
 * 
 * If you find macopt useful, please feel free to make a donation to support
 * David MacKay's research group.
 * 
 * *** THIS CODE COMES WITH NO WARRANTY OF ANY KIND. ***
 */

#include "mexmem.h"
#include "mex.h"
#include "macopt.h"
#include "macoptII.h"

/* A quick note on the ugliness of this. I currently read gradient function name
 * as a string and then give that to Matlab at every callback. This is
 * ridiculous: we should be able to use function handles.
 * Looking at mex.h it looks as though Mathworks, will/does support function
 * handles:
 * 	void mexGetFunctionHandle(void);
 * 	void mexCallMATLABFunction(void);
 * But I certainly can't find any documentation on how to do so. Googling those
 * function names currently doesn't turn up much either.
 */

/* Another note. I was vaguelly aware that numerical recipes use 1-based rather
 * than 0-based arrays, but never really gave much thought to it until this code
 * started seg-faulting all over the floor. So to anyone editing this or writing
 * another front end, remember that any pointer passed to or from macopt refers
 * to the element before the beginning of an array. Otherwise you'll waste as
 * much time as I did working this out. */

/* Renaming: I renamed the files of DJCM's code (when I probably shouldn't have
 * done) because mex is silly and it was the easiest way to make the Matlab
 * function be called "macopt" */

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray*prhs[] )
{
	/*** Variable Declaration ***/
	macopt_args a;               /* MacOpt parameter structure */
	mexCM_args arg;              /* Structure containing args for mex call.
	                                MacOpt kindly passes this around */
	int i;                       /* Looping variable */
	double *initx;               /* Initial condition */
	const int NUM_MANARGS=6;     /* Number of mandatory arguments before those passed to gradfunction start */
	
	/*** Argument parsing and checking ***/
	/* NOTE: if you add more options then change NUM_MANARGS as it's used
	 * later. Also match up the indices where they're used below. Be careful
	 * function x = macopt(start,tol,verbose,itmax,typdist,
	 *                     gradfunction[,gradarg1[,gradarg2[,...]]])
	 * See macopt.m for more on this. */
	if (nrhs < NUM_MANARGS)
		mexErrMsgTxt("Must provide at least: start, tol, verbose, itmax, typdist and gradfunction.\n");
	if (nlhs > 1)
		mexErrMsgTxt("Too many output arguments.\n");
	if(mxGetN(prhs[0])!=1)
		mexErrMsgTxt("Initial point must be a column vector.\n");
	if ((mxGetN(prhs[1])!=1)||(mxGetN(prhs[1])!=1))
		mexErrMsgTxt("Tolerance must be a scalar.\n");
	if ((mxGetN(prhs[2])!=1)||(mxGetN(prhs[2])!=1))
		mexErrMsgTxt("Verbosity level must be a scaler from 0-4.\n");
	if ((mxGetN(prhs[3])!=1)||(mxGetN(prhs[4])!=1))
		mexErrMsgTxt("itmax must be a scalar eg 100.\n");
	if ((mxGetN(prhs[4])!=1)||(mxGetN(prhs[4])!=1))
		mexErrMsgTxt("typdist must be a scalar eg 0.01.\n");
	if(!mxIsChar(prhs[5]))
		mexErrMsgTxt("gradfunction must be string containing gradient function name.\n");
	
	/*** set up macopt parameters ***/
	macopt_defaults(&a);
	a.newitfuncarg=(void*)(&gradient);
	a.tol=*mxGetPr(prhs[1]);
	a.verbose=(int)*mxGetPr(prhs[2]);
	a.itmax=(int)*mxGetPr(prhs[3]);
	a.lastx=*mxGetPr(prhs[4]);
	a.lastx_default=*mxGetPr(prhs[4]);
	
	/*** Set up structure to carry stuff for the gradient function ***/
	arg.dim=mxGetM(prhs[0]);
	arg.fn_len=mxGetN(prhs[5])+1;
	arg.fn=(char *) mxCalloc(arg.fn_len,sizeof(char));
	mxGetString(prhs[5],arg.fn,arg.fn_len);
	/* Create storage for gradient with same dimensions as starting point */
	arg.plhs=(mxArray **)mxCalloc(1,sizeof(mxArray *));
	arg.plhs[0]=mxCreateDoubleMatrix(arg.dim, 1, mxREAL);
	/* Options passed onto gradient function */
	arg.nrhs=nrhs-NUM_MANARGS+1;
	arg.grad_args=(mxArray **)mxCalloc(arg.nrhs,sizeof(mxArray *));
	/* first argument to gradient function arg.grad_args[0] will contain
	 * postion of current point. Fill in rest (if any) with user-provided
	 * parameters for gradient function */
	arg.grad_args[0]=mxCreateDoubleMatrix(arg.dim,1,mxREAL);
	for(i=NUM_MANARGS;i<nrhs;i++)
		arg.grad_args[i-NUM_MANARGS+1]=mxDuplicateArray(prhs[i]);
	
	/*** Run the optimizer ***/
	initx=mxGetPr(prhs[0])-1; /* -1 because nr code is 1-based for
	                             compatibility with Fortran. GARRRRGGHH!! */
	macoptII(initx, arg.dim, gradient, (void*)(&arg), &a);
	
	/*** Clean up ***/
	/* Note that MacOpt cleans itself up */
	mxFree(arg.fn);
	for(i=NUM_MANARGS-1;i<nrhs;i++) 
		mxDestroyArray(arg.grad_args[i-NUM_MANARGS+1]);
	mxFree(arg.grad_args);
	
	/*** Return answer ***/
	plhs[0]=mxDuplicateArray(prhs[0]);
	
	return;
}

/*
 * void gradient(double *x, double *g, void *arg)
 *
 * Gradient callback function. Macopt always calls this which takes care of
 * calling Matlab on its behalf.
 * 
 * I would have defined the gradient callback function thus:
 * 	void gradient(const double *const x, double *const g, void *arg)
 * But it isn't very important and I don't want to change the macopt code.
*/
void gradient(double *x, double *g, void *p_void_arg)
{
	unsigned int i;
	double *ptr;
	
	/* Get some type sanity back */
	mexCM_args *p_arg=(mexCM_args*)p_void_arg;

	/* Copy current location into gradient function's arguments */
	ptr=mxGetPr(p_arg->grad_args[0]);
	for (i=1;i<=p_arg->dim;i++) {
		*ptr=x[i];
		ptr++;
	}
	
	/* Call gradient function and return vector with gradient */
	mexCallMATLAB(1,p_arg->plhs,p_arg->nrhs,p_arg->grad_args,p_arg->fn);
	
	ptr = mxGetPr(p_arg->plhs[0]);
	for (i=1;i<=p_arg->dim;i++) {
		g[i] = *ptr;
		ptr++;
	}
}
